@@ -41,6 +41,7 @@
 #include <sys/un.h>
 #include <sys/types.h>
 #include <string.h>
+#include <stdio.h>
 
 
 #ifndef DEFAULT_BUFFER_SIZE
@@ -118,11 +119,11 @@ static GNetworkConnection *
 create_incoming (GNetworkUnixServer * server, const GValue * server_data, gpointer user_data,
 		 GError ** error)
 {
-  const GNetworkUnixServerCreationData * data = g_value_get_boxed (server_data);
+  const GNetworkUnixServerCreationData *data = g_value_get_boxed (server_data);
 
   return g_object_new (GNETWORK_TYPE_UNIX_CONNECTION,
 		       "connection-type", GNETWORK_CONNECTION_SERVER,
-		       "filename", data->filename, "socket", data->fd, NULL);
+		       "filename", data->filename, "socket", GINT_TO_POINTER (data->fd), NULL);
 }
 
 
@@ -159,6 +160,10 @@ cxn_notify_status_cb (GNetworkConnection * connection, GParamSpec * pspec,
 
   if (status == GNETWORK_CONNECTION_CLOSED)
     {
+      g_signal_handlers_disconnect_by_func (connection, cxn_notify_status_cb, server);
+      g_signal_handlers_disconnect_by_func (connection, cxn_received_cb, server);
+      g_signal_handlers_disconnect_by_func (connection, cxn_sent_cb, server);
+
       server->_priv->connections = g_slist_remove (server->_priv->connections, connection);
       g_object_unref (G_OBJECT (connection));
       g_object_notify (G_OBJECT (server), "connections");
@@ -185,22 +190,24 @@ incoming_handler (GIOChannel * channel, GIOCondition cond, GNetworkUnixServer *
       gint cxn_fd;
       struct sockaddr_un sun;
       socklen_t sun_size;
-      GNetworkUnixServerCreationData creation_data;
+      GNetworkUnixServerCreationData *creation_data;
       GValue value = { 0 };
 
       sun_size = sizeof (struct sockaddr_un);
+      memset (&sun, 0, sun_size);
       cxn_fd = accept (server->_priv->sockfd, (struct sockaddr *) &sun, &sun_size);
 
       /* For whatever reason the incoming connection failed */
       if (cxn_fd < 0)
 	return TRUE;
 
-      creation_data.g_class.g_type = GNETWORK_TYPE_UNIX_SERVER_CREATION_DATA;
-      creation_data.filename = sun.sun_path;
-      creation_data.fd = cxn_fd;
+      creation_data = g_new0 (GNetworkUnixServerCreationData, 1);
+      creation_data->g_class.g_type = GNETWORK_TYPE_UNIX_SERVER_CREATION_DATA;
+      creation_data->filename = g_strdup (server->_priv->filename);
+      creation_data->fd = cxn_fd;
 
       g_value_init (&value, GNETWORK_TYPE_UNIX_SERVER_CREATION_DATA);
-      g_value_set_boxed (&value, &creation_data);
+      g_value_take_boxed (&value, creation_data);
 
       error = NULL;
       cxn = (*server->_priv->create_func) (GNETWORK_SERVER (server), &value, server->_priv->data,
@@ -211,12 +218,13 @@ incoming_handler (GIOChannel * channel, GIOCondition cond, GNetworkUnixServer *
 	{
 	  /* Add it to the list of connections we've got */
 	  server->_priv->connections = g_slist_prepend (server->_priv->connections,
-							g_object_ref (G_OBJECT (cxn)));
+							g_object_ref (cxn));
 
 	  /* Connect signals & tell the world we've got a new incoming connection */
-	  g_signal_connect (cxn, "received", (GCallback) cxn_received_cb, server);
-	  g_signal_connect (cxn, "sent", (GCallback) cxn_sent_cb, server);
-	  g_signal_connect (cxn, "notify::status", (GCallback) cxn_notify_status_cb, server);
+	  g_signal_connect_object (cxn, "received", (GCallback) cxn_received_cb, server, 0);
+	  g_signal_connect_object (cxn, "sent", (GCallback) cxn_sent_cb, server, 0);
+	  g_signal_connect_object (cxn, "notify::status", (GCallback) cxn_notify_status_cb,
+				   server, 0);
 
 	  gnetwork_server_new_connection (GNETWORK_SERVER (server), GNETWORK_CONNECTION (cxn));
 	  g_object_notify (G_OBJECT (server), "connections");
@@ -271,12 +279,15 @@ gnetwork_unix_server_open (GNetworkUnixServer * server)
       switch (errno)
 	{
 	case EMFILE:
-	  error = g_error_new_literal (GNETWORK_SERVER_ERROR, GNETWORK_SERVER_ERROR_TOO_MANY_PROCESSES,
-				       gnetwork_server_strerror (GNETWORK_SERVER_ERROR_TOO_MANY_PROCESSES));
+	  error =
+	    g_error_new_literal (GNETWORK_SERVER_ERROR, GNETWORK_SERVER_ERROR_TOO_MANY_PROCESSES,
+				 gnetwork_server_strerror
+				 (GNETWORK_SERVER_ERROR_TOO_MANY_PROCESSES));
 	  break;
 	case EACCES:
 	  error = g_error_new_literal (GNETWORK_SERVER_ERROR, GNETWORK_SERVER_ERROR_PERMISSIONS,
-				       gnetwork_server_strerror (GNETWORK_SERVER_ERROR_PERMISSIONS));
+				       gnetwork_server_strerror
+				       (GNETWORK_SERVER_ERROR_PERMISSIONS));
 	  break;
 
 	case ENFILE:
@@ -331,6 +342,11 @@ gnetwork_unix_server_open (GNetworkUnixServer * server)
 			       _("The Unix service could not be started because the file \"%s\" "
 				 "could not be created."), server->_priv->filename);
 	  break;
+	case EADDRINUSE:
+	  error = g_error_new (GNETWORK_SERVER_ERROR, GNETWORK_SERVER_ERROR_ALREADY_EXISTS,
+			       _("The Unix service could not be started because the file \"%s\" "
+				 "already exists."), server->_priv->filename);
+	  break;
 
 	default:
 	  error = g_error_new_literal (GNETWORK_SERVER_ERROR, GNETWORK_SERVER_ERROR_INTERNAL,
@@ -343,6 +359,7 @@ gnetwork_unix_server_open (GNetworkUnixServer * server)
 
       shutdown (server->_priv->sockfd, SHUT_RDWR);
       close (server->_priv->sockfd);
+      unlink (server->_priv->filename);
       server->_priv->sockfd = -1;
 
       server->_priv->status = GNETWORK_SERVER_CLOSED;
@@ -362,6 +379,7 @@ gnetwork_unix_server_open (GNetworkUnixServer * server)
 
       shutdown (server->_priv->sockfd, SHUT_RDWR);
       close (server->_priv->sockfd);
+      unlink (server->_priv->filename);
       server->_priv->sockfd = -1;
 
       server->_priv->status = GNETWORK_SERVER_CLOSED;
@@ -370,9 +388,12 @@ gnetwork_unix_server_open (GNetworkUnixServer * server)
     }
 
   server->_priv->channel = g_io_channel_unix_new (server->_priv->sockfd);
-  server->_priv->incoming_id = gnetwork_thread_io_add_watch (server->_priv->channel,
-							     GNETWORK_IO_ANY,
-							     (GIOFunc) incoming_handler, server);
+  server->_priv->incoming_id = gnetwork_thread_io_add_watch_full (server->_priv->channel,
+								  G_PRIORITY_DEFAULT,
+								  GNETWORK_IO_ANY,
+								  (GIOFunc) incoming_handler,
+								  g_object_ref (server),
+								  g_object_unref);
 
   server->_priv->status = GNETWORK_SERVER_OPEN;
   g_object_notify (G_OBJECT (server), "status");
@@ -387,6 +408,7 @@ gnetwork_unix_server_close (GNetworkUnixServer * server)
   if (server->_priv->status <= GNETWORK_SERVER_CLOSED)
     return;
 
+  g_object_freeze_notify (G_OBJECT (server));
   server->_priv->status = GNETWORK_SERVER_CLOSING;
   g_object_notify (G_OBJECT (server), "status");
 
@@ -394,15 +416,28 @@ gnetwork_unix_server_close (GNetworkUnixServer * server)
        g_slist_remove_link (server->_priv->connections, server->_priv->connections))
     {
       if (server->_priv->close_children)
-	gnetwork_connection_close (GNETWORK_CONNECTION (server->_priv->connections->data));
+	{
+	  gnetwork_connection_close (GNETWORK_CONNECTION (server->_priv->connections->data));
+	}
       else
-	g_object_unref (G_OBJECT (server->_priv->connections->data));
+	{
+	  g_signal_handlers_disconnect_by_func (server->_priv->connections->data,
+						cxn_notify_status_cb, server);
+	  g_signal_handlers_disconnect_by_func (server->_priv->connections->data,
+						cxn_received_cb, server);
+	  g_signal_handlers_disconnect_by_func (server->_priv->connections->data,
+						cxn_sent_cb, server);
+	  g_object_unref (server->_priv->connections->data);
+	}
     }
+  g_object_notify (G_OBJECT (server), "connections");
+  g_object_thaw_notify (G_OBJECT (server));
 
   g_io_channel_shutdown (server->_priv->channel, FALSE, NULL);
   g_io_channel_unref (server->_priv->channel);
   server->_priv->channel = NULL;
   server->_priv->sockfd = -1;
+  unlink (server->_priv->filename);
 
   server->_priv->status = GNETWORK_SERVER_CLOSED;
   g_object_notify (G_OBJECT (server), "status");
@@ -428,7 +463,7 @@ gnetwork_unix_server_set_create_func (GNetworkUnixServer * server, GNetworkServe
     {
       server->_priv->create_func = (GNetworkServerCreateFunc) create_incoming;
     }
-  
+
   server->_priv->data = data;
   server->_priv->notify = notify;
 }
@@ -450,7 +485,7 @@ gnetwork_unix_server_server_iface_init (GNetworkServerIface * iface)
 
 static void
 gnetwork_unix_server_set_property (GObject * object, guint property, const GValue * value,
-				  GParamSpec * param_spec)
+				   GParamSpec * param_spec)
 {
   GNetworkUnixServer *server = GNETWORK_UNIX_SERVER (object);
 
@@ -479,7 +514,7 @@ gnetwork_unix_server_set_property (GObject * object, guint property, const GValu
 
 static void
 gnetwork_unix_server_get_property (GObject * object, guint property, GValue * value,
-				  GParamSpec * param_spec)
+				   GParamSpec * param_spec)
 {
   GNetworkUnixServer *server = GNETWORK_UNIX_SERVER (object);
 
@@ -652,7 +687,8 @@ gnetwork_unix_server_creation_data_get_type (void)
     {
       type = g_boxed_type_register_static ("GNetworkUnixServerCreationData",
 					   (GBoxedCopyFunc) gnetwork_unix_server_creation_data_dup,
-					   (GBoxedFreeFunc) gnetwork_unix_server_creation_data_free);
+					   (GBoxedFreeFunc)
+					   gnetwork_unix_server_creation_data_free);
     }
 
   return type;
